/*
 * DuneApp.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <errno.h>
#include <unistd.h>
#include "stdafx.h"
#include "swt.h"

#include "DuneApp.h"
#include "Scene.h"
#include "MainWindow.h"
#include "URL.h"
#include "EulerAngles.h"
#include "resource.h"

DuneApp *TheApp = NULL;
#define MAX_RECENT_FILES	4

DuneApp::DuneApp()
{
    TheApp = this;
    swSetDefaultIcon(IDI_DUNE_ICON);
    _prefs = swLoadPreferences("DuneSoft", "dune");
    _showAllFields = GetBoolPreference("ShowAllFields", false);
    _rotationOrder = GetIntPreference("RotationOrder", EulOrdXYZs);
    _rotationTitle = "XYZs";
    _handleMode = (HandleMode) GetIntPreference("HandleMode", HM_SELECTED);
    _browser = swBrowserInit(_prefs);

    int n = GetIntPreference("RecentFiles", 0);
    for (int i = 0; i < n; i++) {
	char		buf[1024];
	const char     *filename;
	sprintf(buf, "RecentFile%d", i);
	filename = TheApp->GetPreference(buf, "");
	_recentFiles.append(filename);
    }
}

void
DuneApp::SaveRecentFileList()
{
    int		i, n = _recentFiles.size();

    SetIntPreference("RecentFiles", n);
    for (i = 0; i < n; i++) {
	char		buf[1024];
	sprintf(buf, "RecentFile%d", i);
	SetPreference(buf, _recentFiles[i]);
    }
}

void
DuneApp::OnFileNew()
{
    Scene	*scene = new Scene();
    int		width = GetIntPreference("WindowWidth", 800);
    int		height = GetIntPreference("WindowHeight", 600);
    SWND mainWnd = swCreateMainWindow("Dune", 0, 0, width, height);
    MainWindow	*window = new MainWindow(scene, mainWnd);
    _windows.append(window);
}

bool
DuneApp::OpenFile(const char *path)
{
    FILE       *file;

    if (!path[0] || !(file = fopen(path, "r"))) {
	return false;
    }

    Scene	*scene = new Scene();
    const char *errors = scene->parse(file);
    fclose(file);
    if (errors[0]) {
	swMessageBox(NULL, errors, "Parse Errors", SW_MB_OK, SW_MB_WARNING);
	swDebugf(errors);
	return false;
    }

    URL		url;

    url.FromPath(path);

    scene->setURL(url);
    scene->setPath(path);
    scene->setSelection(scene->getRoot());
    scene->UpdateViews(NULL, UPDATE_SELECTION);
    scene->UpdateViews(NULL, UPDATE_ALL);
    AddToRecentFiles(path);
    int		width = GetIntPreference("WindowWidth", 800);
    int		height = GetIntPreference("WindowHeight", 600);
    SWND mainWnd = swCreateMainWindow("Dune", 0, 0, width, height);
    MainWindow	*window = new MainWindow(scene, mainWnd);
    _windows.append(window);
    return true;
}

void DuneApp::OnFileClose(MainWindow *window)
{
    if (!window->SaveModified()) {
	return;
    }

    _windows.remove(_windows.find(window));
    delete window;
    if (_windows.size() == 0) {
	Exit();
    }
}

void DuneApp::OnFileExit()
{
    List<MainWindow *>::Iterator *i;
    for (i = _windows.first(); i; i = i->next()) {
	if (!i->item()->SaveModified()) return;
    }
    for (i = _windows.first(); i; i = i->next()) {
	delete i->item();
    }
    Exit();
}

void DuneApp::Exit()
{
    SetBoolPreference("ShowAllFields", _showAllFields);
    SetIntPreference("RotationOrder", _rotationOrder);
    SetIntPreference("HandleMode", _handleMode);
    SaveRecentFileList();
    swSavePreferences(_prefs);
    swBrowserShutdown(_browser);
    swDeletePreferences(_prefs);
    swExit();
}

void DuneApp::SetShowAllFields(bool showAllFields)
{
    if (_showAllFields != showAllFields) {
	_showAllFields = showAllFields;
	UpdateAllWindows();
    }
}

void DuneApp::SetRotationOrder(int rotationOrder)
{
    if (_rotationOrder != rotationOrder) {
	_rotationOrder = rotationOrder;
	UpdateAllWindows();
    }
}

void DuneApp::SetHandleMode(HandleMode handleMode)
{
    if (_handleMode != handleMode) {
	_handleMode = handleMode;
	UpdateAllWindows();
    }
}

void DuneApp::UpdateAllWindows()
{
    for (List<MainWindow *>::Iterator *i = _windows.first(); i; i = i->next()) {
	i->item()->GetScene()->UpdateViews(NULL, UPDATE_ALL);
    }
}

bool
DuneApp::GetBoolPreference(const char *key, bool defaultValue)
{
    return swGetIntPreference(_prefs, key, defaultValue) != 0;
}

int
DuneApp::GetIntPreference(const char *key, int defaultValue)
{
    return swGetIntPreference(_prefs, key, defaultValue);
}

const char *
DuneApp::GetPreference(const char *key, const char *defaultValue)
{
    return swGetPreference(_prefs, key, defaultValue);
}

void
DuneApp::SetBoolPreference(const char *key, bool value)
{
    swSetIntPreference(_prefs, key, value ? 1 : 0);
}

void
DuneApp::SetIntPreference(const char *key, int value)
{
    swSetIntPreference(_prefs, key, value);
}

void
DuneApp::SetPreference(const char *key, const char *value)
{
    swSetPreference(_prefs, key, value);
}

int
DuneApp::GetNumRecentFiles() const
{
    return _recentFiles.size();
}

const String &
DuneApp::GetRecentFile(int index) const
{
    return _recentFiles[index];
}

void
DuneApp::AddToRecentFiles(const String &filename)
{
    int i = _recentFiles.find(filename);

    if (i >= 0) {
	_recentFiles.remove(i);
    }
    int		n = MIN(_recentFiles.size(), MAX_RECENT_FILES-1);
    for (i = n; i > 0; i--) {
	_recentFiles[i] = _recentFiles[i-1];
    }
    _recentFiles[0] = filename;
}
