/*
 * NodeElevationGrid.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <GL/gl.h>
#include "stdafx.h"

#include "NodeElevationGrid.h"
#include "MFColor.h"
#include "MFFloat.h"
#include "MFInt32.h"
#include "MFVec2f.h"
#include "MFVec3f.h"
#include "SFBool.h"
#include "SFFloat.h"
#include "SFInt32.h"
#include "SFNode.h"
#include "Scene.h"
#include "Mesh.h"
#include "RenderState.h"

#define HEIGHT(i, j)	(height[(i) + (j) * xDimension])

ProtoElevationGrid::ProtoElevationGrid(Scene *scene)
  : Proto(scene, "ElevationGrid")
{
    addEventIn(MFFLOAT, "set_height");
    addExposedField(SFNODE, "color", new SFNode(NULL), NODE_COLOR);
    addExposedField(SFNODE, "normal", new SFNode(NULL), NODE_NORMAL);
    addExposedField(SFNODE, "texCoord", new SFNode(NULL), NODE_TEXTURE_COORDINATE);
    addField(MFFLOAT, "height", new MFFloat());
    addField(SFBOOL, "ccw", new SFBool(true));
    addField(SFBOOL, "colorPerVertex", new SFBool(true));
    addField(SFFLOAT, "creaseAngle", new SFFloat(0.0f), new SFFloat(0.0f));
    addField(SFBOOL, "normalPerVertex", new SFBool(true));
    addField(SFBOOL, "solid", new SFBool(true));
    addField(SFINT32, "xDimension", new SFInt32(0), new SFInt32(0));
    addField(SFFLOAT, "xSpacing", new SFFloat(1.0f), new SFInt32(0));
    addField(SFINT32, "zDimension", new SFInt32(0), new SFInt32(0));
    addField(SFFLOAT, "zSpacing", new SFFloat(1.0f), new SFInt32(0));
}

NodeElevationGrid::NodeElevationGrid(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _mesh = NULL;
    _meshDirty = true;
}

NodeElevationGrid::~NodeElevationGrid()
{
    delete _mesh;
}

void
NodeElevationGrid::draw()
{
    if (_meshDirty) {
	createMesh();
	_meshDirty = false;
    }

    if (!_mesh) return;
    
    _mesh->draw();
}

void
NodeElevationGrid::createMesh()
{
    Node	 *color = ((SFNode *) getField(0))->getValue();
    Node	 *normal = ((SFNode *) getField(1))->getValue();
    Node	 *texCoord = ((SFNode *) getField(2))->getValue();
    const float  *height = ((MFFloat *) getField(3))->getValues();
    bool	  ccw = ((SFBool *) getField(4))->getValue();
    bool	  colorPerVertex = ((SFBool *) getField(5))->getValue();
    float	  creaseAngle = ((SFFloat *) getField(6))->getValue();
    bool	  normalPerVertex = ((SFBool *) getField(7))->getValue();
    bool	  solid = ((SFBool *) getField(8))->getValue();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();

    if (xDimension == 0 || zDimension == 0) return;

    MFVec3f    *normals = normal ? (MFVec3f *) normal->getField(0) : NULL;
    MFColor    *colors = color ? (MFColor *) color->getField(0) : NULL;
    MFVec2f    *texCoords = texCoord ? (MFVec2f *) texCoord->getField(0) : NULL;

    int		size = xDimension * zDimension;
    Vec3f      *vertices = new Vec3f[size];
    Vec2f      *tc = new Vec2f[size];
    int		index = 0;
    int		i, j;

    for (j = 0; j < zDimension; j++) {
	for (i = 0; i < xDimension; i++) {
	    vertices[index] = Vec3f(i * xSpacing, HEIGHT(i, j), j * zSpacing);
	    tc[index] = Vec2f(i / (xDimension-1.0f), j / (zDimension-1.0f));
	    index++;
	}
    }
    index = 0;
    int        *ci = new int[size * 8];

    for (j = 0; j < zDimension-1; j++) {
	for (i = 0; i < xDimension-1; i++) {
	    ci[index++] = j * xDimension + i;
	    ci[index++] = (j+1) * xDimension + (i+1);
	    ci[index++] = j * xDimension + (i+1);
	    ci[index++] = -1;
	    ci[index++] = j * xDimension + i;
	    ci[index++] = (j+1) * xDimension + i;
	    ci[index++] = (j+1) * xDimension + (i+1);
	    ci[index++] = -1;
	}
    }
    MFVec3f *v = new MFVec3f((float *) vertices, size * 3);
    MFInt32 *coordIndex = new MFInt32(ci, index);
    if (!texCoords) texCoords = new MFVec2f((float *) tc, size * 2);

    delete _mesh;
    _mesh = new Mesh(v, coordIndex, normals, NULL,
		     colors, NULL, texCoords, NULL, creaseAngle, ccw, solid);
}

void
NodeElevationGrid::drawHandles()
{
    const float  *height = ((MFFloat *) getField(3))->getValues();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();
    RenderState	  state;

    if (xDimension < 2 || zDimension < 2) return;

    if (((MFFloat *) getField(3))->getSize() != xDimension * zDimension)
	return;

    glDisable(GL_LIGHTING);
    glPushName(0);
    glColor3f(1.0f, 1.0f, 1.0f);
    state.startDrawHandles();
    for (int j = 0; j < zDimension; j++) {
	for (int i = 0; i < xDimension; i++) {
	    glLoadName(j * xDimension + i);
	    state.drawHandle(Vec3f(i * xSpacing, HEIGHT(i, j), j * zSpacing));
	}
    }
    state.endDrawHandles();
    glPopName();
    glEnable(GL_LIGHTING);
}

void
NodeElevationGrid::setField(int index, FieldValue *value)
{
    _meshDirty = true;
    Node::setField(index, value);
}

Vec3f
NodeElevationGrid::getHandle(int handle, int *constraint, int *field)
{
    const float  *height = ((MFFloat *) getField(3))->getValues();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();
    float	  x = (handle % xDimension) * xSpacing;
    float	  y = height[handle];
    float	  z = handle / xDimension * zSpacing;

    *field = 3;
    *constraint = CONSTRAIN_Y;
    return Vec3f(x, y, z);
}

void
NodeElevationGrid::setHandle(int handle, const Vec3f &v)
{
    MFFloat	*oldValue = (MFFloat *) getField(3);
    MFFloat	*newValue = new MFFloat(*oldValue);

    newValue->setValue(handle, v.y);

    _scene->setField(this, 3, newValue);
}
