/*
 * NodeIndexedFaceSet.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <GL/gl.h>
#include "stdafx.h"

#include "NodeIndexedFaceSet.h"
#include "Proto.h"
#include "FieldValue.h"
#include "MFInt32.h"
#include "SFBool.h"
#include "SFFloat.h"
#include "SFNode.h"
#include "SFInt32.h"
#include "Node.h"
#include "MFVec2f.h"
#include "MFVec3f.h"
#include "MFColor.h"
#include "Mesh.h"
#include "Face.h"
#include "Vec3f.h"


ProtoIndexedFaceSet::ProtoIndexedFaceSet(Scene *scene)
  : Proto(scene, "IndexedFaceSet")
{
    addEventIn(MFINT32, "set_colorIndex");
    addEventIn(MFINT32, "set_coordIndex");
    addEventIn(MFINT32, "set_normalIndex");
    addEventIn(MFINT32, "set_texCoordIndex");
    addExposedField(SFNODE, "color", new SFNode(NULL), NODE_COLOR);
    addExposedField(SFNODE, "coord", new SFNode(NULL), NODE_COORDINATE);
    addExposedField(SFNODE, "normal", new SFNode(NULL), NODE_NORMAL);
    addExposedField(SFNODE, "texCoord", new SFNode(NULL), NODE_TEXTURE_COORDINATE);
    addField(SFBOOL, "ccw", new SFBool(true));
    addField(MFINT32, "colorIndex", new MFInt32(), new SFInt32(-1));
    addField(SFBOOL, "colorPerVertex", new SFBool(true));
    addField(SFBOOL, "convex", new SFBool(true));
    addField(MFINT32, "coordIndex", new MFInt32(), new SFInt32(-1));
    addField(SFFLOAT, "creaseAngle", new SFFloat(0.0), new SFFloat(0.0f));
    addField(MFINT32, "normalIndex", new MFInt32(), new SFInt32(-1));
    addField(SFBOOL, "normalPerVertex", new SFBool(true));
    addField(SFBOOL, "solid", new SFBool(true));
    addField(MFINT32, "texCoordIndex", new MFInt32(), new SFInt32(-1));
}

NodeIndexedFaceSet::NodeIndexedFaceSet(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _mesh = NULL;
    _meshDirty = true;
}

NodeIndexedFaceSet::~NodeIndexedFaceSet()
{
    delete _mesh;
}

void
NodeIndexedFaceSet::setField(int index, FieldValue *value)
{
    _meshDirty = true;
    Node::setField(index, value);
}

void
NodeIndexedFaceSet::draw()
{
    if (_meshDirty) {
	createMesh();
	_meshDirty = false;
    }

    if (!_mesh) return;

    _mesh->draw();
}

void
NodeIndexedFaceSet::createMesh()
{
    Node       *color = ((SFNode *) getField(0))->getValue();
    Node       *coord = ((SFNode *) getField(1))->getValue();
    Node       *normal = ((SFNode *) getField(2))->getValue();
    Node       *texCoord = ((SFNode *) getField(3))->getValue();
    bool	ccw = ((SFBool *) getField(4))->getValue();
    MFInt32    *colorIndex = (MFInt32 *) getField(5);
    bool	colorPerVertex = ((SFBool *) getField(6))->getValue();
    bool	convex = ((SFBool *) getField(7))->getValue();
    MFInt32    *coordIndex = (MFInt32 *) getField(8);
    float	creaseAngle = ((SFFloat *) getField(9))->getValue();
    MFInt32    *normalIndex = (MFInt32 *) getField(8);
    bool	solid = ((SFBool *) getField(12))->getValue();
    MFInt32    *texCoordIndex = (MFInt32 *) getField(13);
   
    if (!coord || coord->getField(0)->getType() != MFVEC3F)
	return;

    MFVec3f		*coords = (MFVec3f *) coord->getField(0);
    MFVec3f		*normals = NULL;
    MFColor		*colors = NULL;
    MFVec2f		*texCoords = NULL;

    if (normal) {
	normals = (MFVec3f *) normal->getField(0);
    }
    if (color) {
	colors = (MFColor *) color->getField(0);
    }
    if (texCoord) {
	texCoords = (MFVec2f *) texCoord->getField(0);
    }
    if (colorIndex->getSize() != coordIndex->getSize()) {
	colorIndex = coordIndex;
    }
    if (texCoordIndex->getSize() != coordIndex->getSize()) {
	texCoordIndex = coordIndex;
    }
    if (normalIndex->getSize() != coordIndex->getSize()) {
	normalIndex = coordIndex;
    }
    delete _mesh;
    _mesh = new Mesh(coords, coordIndex, normals, normalIndex,
		     colors, colorIndex, texCoords, texCoordIndex,
		     creaseAngle, ccw, solid);
}

