/*
 * NodeNurbsSurface.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <GL/gl.h>
#include <assert.h>
#include <alloca.h>

#include "stdafx.h"
#include "NodeNurbsSurface.h"
#include "Scene.h"
#include "FieldValue.h"
#include "SFInt32.h"
#include "Mesh.h"
#include "MFFloat.h"
#include "MFInt32.h"
#include "MFVec2f.h"
#include "MFVec3f.h"
#include "SFNode.h"
#include "SFBool.h"
#include "SFVec3f.h"
#include "Vec2f.h"
#include "Vec3f.h"
#include "RenderState.h"

ProtoNurbsSurface::ProtoNurbsSurface(Scene *scene)
  : Proto(scene, "NurbsSurface")
{
    addField(SFINT32, "uDimension", new SFInt32(0), new SFInt32(0));
    addField(SFINT32, "vDimension", new SFInt32(0), new SFInt32(0));
    addField(MFFLOAT, "uKnot", new MFFloat());
    addField(MFFLOAT, "vKnot", new MFFloat());
    addField(SFINT32, "uOrder", new SFInt32(3), new SFInt32(2));
    addField(SFINT32, "vOrder", new SFInt32(3), new SFInt32(2));
    addExposedField(MFVEC3F, "controlPoint", new MFVec3f());
    addExposedField(MFFLOAT, "weight", new MFFloat(), new SFFloat(0.0f));
    addExposedField(SFINT32, "uTessellation", new SFInt32(0));
    addExposedField(SFINT32, "vTessellation", new SFInt32(0));
    addExposedField(SFNODE, "texCoord", new SFNode(NULL), NODE_COORDINATE);
    addField(SFBOOL, "ccw", new SFBool(true));
    addField(SFBOOL, "solid", new SFBool(true));
}

NodeNurbsSurface::NodeNurbsSurface(Scene *scene, Proto *proto)
  : Node(scene, proto)
{
    _mesh = NULL;
    _meshDirty = true;
}

NodeNurbsSurface::~NodeNurbsSurface()
{
    delete _mesh;
}

void
NodeNurbsSurface::draw()
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }

    if (!_mesh) return;

    _mesh->draw();
}

void
NodeNurbsSurface::createMesh()
{
    int		uDimension = ((SFInt32 *) getField(0))->getValue();
    int		vDimension = ((SFInt32 *) getField(1))->getValue();
    MFFloat    *uKnot = (MFFloat *) getField(2);
    MFFloat    *vKnot = (MFFloat *) getField(3);
    int		uOrder = ((SFInt32 *) getField(4))->getValue();
    int		vOrder = ((SFInt32 *) getField(5))->getValue();
    MFVec3f    *controlPoint = (MFVec3f *) getField(6);
    MFFloat    *weight = (MFFloat *) getField(7);
    int		uTess = ((SFInt32 *) getField(8))->getValue();
    int		vTess = ((SFInt32 *) getField(9))->getValue();
    Node       *texCoord = ((SFNode *) getField(10))->getValue();
    bool	ccw = ((SFBool *) getField(11))->getValue();
    bool	solid = ((SFBool *) getField(12))->getValue();
    float      *weights = NULL;
    int		i, j;

    if (uDimension == 0 || vDimension == 0) return;

    if (uKnot->getSize() != uOrder + uDimension
     || vKnot->getSize() != vOrder + vDimension)
	return;

    if (controlPoint->getSize() != uDimension * vDimension * 3)
	return;

    if (weight->getSize() == 0) {
	weights = new float[uDimension * vDimension];
	for (i = 0; i < uDimension * vDimension; i++) {
	    weights[i] = 1.0f;
	}
    } else if (weight->getSize() != uDimension * vDimension) {
	return;
    }

    MFVec2f *texCoords = texCoord ? (MFVec2f *) texCoord->getField(0) : NULL;

    if (uTess <= 0) uTess = 32;
    if (vTess <= 0) vTess = 32;
    int		size = (uTess + 1) * (vTess + 1);
    Vec3f      *tess = new Vec3f[size];
    Vec3f      *normals = new Vec3f[size];

    const float	*uKnots = uKnot->getValues();
    const float *vKnots = vKnot->getValues();
    float	uInc = (uKnots[uKnot->getSize()-1] - uKnots[0]) / uTess;
    float 	vInc = (vKnots[vKnot->getSize()-1] - vKnots[0]) / vTess;

    int		index = 0;
    float	u, v;
    const float *w = weights ? weights : weight->getValues();
    for (j = 0, v = vKnots[0]; j <= vTess; j++, v += vInc) {
	for (i = 0, u = uKnots[0]; i <= uTess; i++, u += uInc) {
	    tess[index] = surfacePoint(uDimension, uOrder, uKnots,
				       vDimension, vOrder, vKnots,
				       (const Vec3f *) controlPoint->getValues(),
				       w, u, v, normals[index]);
	    if (!ccw) normals[index] = -normals[index];
	    index++;
	}
    }

    index = 0;
    Vec2f	       *tc = new Vec2f[size];
    float		uinv = 1.0f / uTess;
    float		vinv = 1.0f / vTess;
    for (j = 0; j <= vTess; j++) {
	for (i = 0; i <= uTess; i++) {
	    tc[index++] = Vec2f(i * uinv, j * vinv);
	}
    }
    int		       *ci = new int[size * 5];
    index = 0;
    for (j = 0; j < vTess; j++) {
        for (i = 0; i < uTess; i++) {
	    ci[index++] = j * (uTess+1) + i;
	    ci[index++] = j * (uTess+1) + (i+1);
	    ci[index++] = (j+1) * (uTess+1) + (i+1);
	    ci[index++] = (j+1) * (uTess+1) + i;
            ci[index++] = -1;
        }
    }

    MFVec3f *vertices = new MFVec3f((float *) tess, size);
    MFVec3f *normal = new MFVec3f((float *) normals, size);
    MFInt32 *coordIndex = new MFInt32(ci, index);
    if (!texCoords) texCoords = new MFVec2f((float *) tc, size * 2);

    delete _mesh;
    _mesh = new Mesh(vertices, coordIndex, normal, coordIndex,
                     NULL, NULL, texCoords, coordIndex, 1.57f, ccw, solid);
    delete [] weights;
}

void
NodeNurbsSurface::drawHandles()
{
    int		uDimension = ((SFInt32 *) getField(0))->getValue();
    int		vDimension = ((SFInt32 *) getField(1))->getValue();
    MFVec3f    *controlPoint = (MFVec3f *) getField(6);
    MFFloat    *weight = (MFFloat *) getField(7);
    RenderState	state;

    if (controlPoint->getSize() != uDimension * vDimension * 3)
	return;

    if (weight->getSize() != uDimension * vDimension) {
	return;
    }

    glPushName(uDimension * vDimension + 1);
    glDisable(GL_LIGHTING);
    glColor3f(1.0f, 1.0f, 1.0f);
    for (int i = 0; i < uDimension; i++) {
	glBegin(GL_LINE_STRIP);
	for (int j = 0; j < vDimension; j++) {
	    const float *v = controlPoint->getValue(i + j*uDimension);
	    float	w = weight->getValue(i + j*uDimension);
	    glVertex3f(v[0] / w, v[1] / w, v[2] / w);
	}
	glEnd();
    }
    for (int j = 0; j < vDimension; j++) {
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i < uDimension; i++) {
	    const float *v = controlPoint->getValue(i + j*uDimension);
	    float	w = weight->getValue(i + j*uDimension);
	    glVertex3f(v[0] / w, v[1] / w, v[2] / w);
	}
	glEnd();
    }
    state.startDrawHandles();
    for (int ci = 0; ci < uDimension * vDimension; ci++) {
	glLoadName(ci);
	state.drawHandle(
		Vec3f(controlPoint->getValue(ci)) / weight->getValue(ci));
    }
    state.endDrawHandles();
    glPopName();
    glEnable(GL_LIGHTING);
}

Vec3f
NodeNurbsSurface::getHandle(int handle, int *constraint,
			    int *field)
{
    MFVec3f    *controlPoint = (MFVec3f *) getField(6);
    MFFloat    *weight = (MFFloat *) getField(7);

    *constraint = CONSTRAIN_NONE;
    *field = 6;

    if (handle >= 0 && handle < controlPoint->getSize()) {
	return Vec3f(controlPoint->getValue(handle)) / weight->getValue(handle);
    } else {
	return Vec3f(0.0f, 0.0f, 0.0f);
    }
}

void
NodeNurbsSurface::setHandle(int handle, const Vec3f &v)
{
    MFVec3f    *oldValue = (MFVec3f *) getField(6);
    MFVec3f    *newValue = new MFVec3f(*oldValue);
    MFFloat    *weight = (MFFloat *) getField(7);
    Vec3f	v2 = v * weight->getValue(handle);	 

    if (handle >= 0 && handle < oldValue->getSize() / 3) {
	_meshDirty = true;
	newValue->setValue(handle * 3, v2.x);
	newValue->setValue(handle * 3+1, v2.y);
	newValue->setValue(handle * 3+2, v2.z);
	_scene->setField(this, 6, newValue);
    }
}

void
NodeNurbsSurface::setField(int index, FieldValue *value)
{
    _meshDirty = true;
    Node::setField(index, value);
}

int
NodeNurbsSurface::findSpan(int dimension, int order, float u,
			   const float knots[])
{
    int		low, mid, high;
    int		n = dimension + order - 1;

    if (u >= knots[n]) {
	return n - order;
    }
    low = order - 1;	high = n - order + 1;

    mid = (low + high) / 2;

    while (u < knots[mid] || u >= knots[mid+1]) {
	if (u < knots[mid]) {
	    high = mid;
	} else {
	    low = mid;
	}
	mid = (low+high)/2;
    }
    return mid;
}

void
NodeNurbsSurface::basisFuns(int span, float u, int order,
			    const float knots[], float basis[],
			    float deriv[])
{
    float      *left = (float *) alloca(order * sizeof(float));
    float      *right = (float *) alloca(order * sizeof(float));

    basis[0] = 1.0f;
    for (int j = 1; j < order; j++) {
	left[j] = u - knots[span+1-j];
	right[j] = knots[span+j]-u;
	float saved = 0.0f, dsaved = 0.0f;
	for (int r = 0; r < j; r++) {
	    float temp = basis[r] / (right[r+1] + left[j-r]);
	    basis[r] = saved + right[r+1] * temp;
	    deriv[r] = dsaved - j * temp;
	    saved = left[j-r] * temp;
	    dsaved = j * temp;
	}
	basis[j] = saved;
	deriv[j] = dsaved;
    }
}

Vec3f
NodeNurbsSurface::surfacePoint(int uDimension, int uOrder,
			       const float uKnots[],
			       int vDimension, int vOrder,
			       const float vKnots[],
			       const Vec3f controlPoints[],
			       const float weight[], float u, float v,
			       Vec3f &normal)
{
    float      *uBasis = (float *) alloca(uOrder * sizeof(float));
    float      *vBasis = (float *) alloca(vOrder * sizeof(float));
    float      *uDeriv = (float *) alloca(uOrder * sizeof(float));
    float      *vDeriv = (float *) alloca(uOrder * sizeof(float));

    int		uSpan = findSpan(uDimension, uOrder, u, uKnots);
    int		vSpan = findSpan(vDimension, vOrder, v, vKnots);

    basisFuns(uSpan, u, uOrder, uKnots, uBasis, uDeriv);
    basisFuns(vSpan, v, vOrder, vKnots, vBasis, vDeriv);

    int		uBase = uSpan-uOrder+1;
    int		vBase = vSpan-vOrder+1;

    int		index = vBase*uDimension + uBase;
    Vec3f S(0.0f, 0.0f, 0.0f), du(0.0f, 0.0f, 0.0f), dv(0.0f, 0.0f, 0.0f);
    float w = 0.0f, duw = 0.0f, dvw = 0.0f;
    for (int j = 0; j < vOrder; j++) {
	for (int i = 0; i < uOrder; i++) {
	    float	gain = uBasis[i] * vBasis[j];
	    float	dugain = uDeriv[i] * vBasis[j];
	    float	dvgain = uBasis[i] * vDeriv[j];
	    S += controlPoints[index] * gain;
	    w += weight[index] * gain;
	    du += controlPoints[index] * dugain;
	    dv += controlPoints[index] * dvgain;
	    duw += weight[index] * dugain;
	    dvw += weight[index] * dvgain;
	    index++;
	}
	index += uDimension - uOrder;
    }
    S = S / w;
    Vec3f un = (du - S * duw) / w;
    Vec3f vn = (dv - S * dvw) / w;
    normal = un.cross(vn);
    normal.normalize();
    return S;
}
