/*
 * NodePointLight.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "stdafx.h"

#include "NodePointLight.h"
#include "Scene.h"
#include "Proto.h"
#include "FieldValue.h"
#include "SFFloat.h"
#include "SFVec3f.h"
#include "SFColor.h"
#include "SFBool.h"

enum {
    LOCATION,
};

ProtoPointLight::ProtoPointLight(Scene *scene)
  : Proto(scene, "PointLight")
{
    addExposedField(SFFLOAT, "ambientIntensity", new SFFloat(0.0f),
    		    new SFFloat(0.0f), new SFFloat(1.0f));
    addExposedField(SFVEC3F, "attenuation", new SFVec3f(1.0f, 0.0f, 0.0f),
    		    new SFFloat(0.0f));
    addExposedField(SFCOLOR, "color", new SFColor(1.0f, 1.0f, 1.0f));
    addExposedField(SFFLOAT, "intensity", new SFFloat(1.0f), 
    		    new SFFloat(0.0f), new SFFloat(1.0f));
    addExposedField(SFVEC3F, "location", new SFVec3f(0.0f, 0.0f, 0.0f));
    addExposedField(SFBOOL, "on", new SFBool(true));
    addExposedField(SFFLOAT, "radius", new SFFloat(100.0f), new SFFloat(0.0f));
}

NodePointLight::NodePointLight(Scene *scene, Proto *def)
  : Node(scene, def)
{
}

void
NodePointLight::preDraw()
{
    float	    ambientIntensity = ((SFFloat *) getField(0))->getValue();
    const float	   *attenuation = ((SFVec3f *) getField(1))->getValue();
    const float	   *color = ((SFColor *) getField(2))->getValue();
    float	    intensity = ((SFFloat *) getField(3))->getValue();
    const float	   *location = ((SFVec3f *) getField(4))->getValue();
    int		    on = ((SFBool *) getField(5))->getValue();
//    float	    radius = ((SFFloat *) getField(6))->getValue();

    if (on) {
	float ambientColor[4], diffuseColor[4];
	float pos[4];
	for (int i = 0; i < 3; i++) {
	    ambientColor[i] = color[i] * ambientIntensity;
	    diffuseColor[i] = color[i] * intensity;
	    pos[i] = location[i];
	}
	ambientColor[3] = diffuseColor[3] = 1.0f;
	pos[3] = 1.0f;

	_light = (GLenum) _scene->allocateLight();

	glLightfv(_light, GL_AMBIENT, ambientColor);
	glLightfv(_light, GL_DIFFUSE, diffuseColor);
	glLightfv(_light, GL_POSITION, pos);
	glLightfv(_light, GL_SPECULAR, diffuseColor);
	glLightf(_light, GL_SPOT_CUTOFF, 180.0f);
	glLightf(_light, GL_SPOT_EXPONENT, 0.0f);
	glLightf(_light, GL_CONSTANT_ATTENUATION, attenuation[0]);
	glLightf(_light, GL_LINEAR_ATTENUATION, attenuation[1]);
	glLightf(_light, GL_QUADRATIC_ATTENUATION, attenuation[2]);
	glEnable(_light);
    }
}

void
NodePointLight::drawHandles()
{
    const float	   *location = ((SFVec3f *) getField(4))->getValue();

    glPushMatrix();
    glTranslatef(location[0], location[1], location[2]);

    glPushName(LOCATION);

    GLUquadricObj   *obj = gluNewQuadric();

    const float	   *color = ((SFColor *) getField(2))->getValue();
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);
    gluSphere(obj, 0.05f, 8, 4);

    glPopMatrix();

    gluDeleteQuadric(obj);
}

Vec3f
NodePointLight::getHandle(int handle, int *constraint, int *field)
{
    const float	   *location = ((SFVec3f *) getField(4))->getValue();

    *field = 4;
    *constraint = CONSTRAIN_NONE;
    return location;
}

void
NodePointLight::setHandle(int handle, const Vec3f &v)
{
    _scene->setField(this, 4, new SFVec3f(v));
}
