/*
 * SFRotation.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <math.h>
#include "stdafx.h"

#include "SFRotation.h"
#include "EulerAngles.h"

SFRotation::SFRotation(float x, float y, float z, float w)
  : _quaternion(Vec3f(x, y, z), w)
{
    _value[0] = x; _value[1] = y; _value[2] = z; _value[3] = w;
    _eulerDirty = true;
}

SFRotation::SFRotation(const float *value)
  : _quaternion(Vec3f(value[0], value[1], value[2]), value[3])
{
    _value[0] = value[0];
    _value[1] = value[1];
    _value[2] = value[2];
    _value[3] = value[3];
    _eulerDirty = true;
}

SFRotation::SFRotation(Vec3f angles, int mode)
{
    setEulerAngles(angles, mode);
}

SFRotation::SFRotation(const Quaternion &q)
  : _quaternion(q)
{
    float rlen = (float) sqrt(q.x * q.x + q.y * q.y + q.z * q.z);
    if (rlen > 0.0f) rlen = 1.0f / rlen; else rlen = 1.0f;
    _value[0] = q.x * rlen;
    _value[1] = q.y * rlen;
    _value[2] = q.z * rlen;
    _value[3] = 2.0f * (float) acos(q.w);
    _eulerDirty = true;
}

SFRotation::SFRotation(const Matrix &mat)
{
    HMatrix	hmat;
    mat.getValue(hmat);
    EulerAngles	a = Eul_FromHMatrix(hmat, EulOrdXYZs);

    setEulerAngles(Vec3f(a.x, a.y, a.z), EulOrdXYZs);
}

bool
SFRotation::equals(const FieldValue *value) const
{
    if (value->getType() == SFROTATION) {
	for (int i = 0; i < 4; i++)
	    if (((SFRotation *) value)->getValue()[i] != _value[i])
		return false;
	return true;
    }
    return false;
}

void SFRotation::write(FILE *f, int /* indent */) const
{
    fprintf(f, "%g %g %g %g\n", _value[0], _value[1], _value[2], _value[3]);
}

Vec3f SFRotation::getEulerAngles(int order)
{
    if (_eulerDirty || _eulerOrder != order) {
	Vec3f		v(_value[0], _value[1], _value[2]);
	Quaternion	q(v, _value[3]);
	EulerAngles	a;
	a = Eul_FromQuat(q, order);
	_euler = Vec3f(RAD2DEG(a.x), RAD2DEG(a.y), RAD2DEG(a.z));
	_eulerOrder = order;
	_eulerDirty = false;
    }
    return _euler;
}

const Quaternion &
SFRotation::getQuat() const
{
    return _quaternion;
}

void
SFRotation::setEulerAngles(Vec3f angles, int order)
{
    Quat	q;
    EulerAngles	a;

    a.x = DEG2RAD(angles.x);
    a.y = DEG2RAD(angles.y);
    a.z = DEG2RAD(angles.z);
    a.w = (float) order;

    q = Eul_ToQuat(a);
    _value[0] = q.x;
    _value[1] = q.y;
    _value[2] = q.z;
    _value[3] = 2.0f * (float) acos(q.w);
    _quaternion = Quaternion(q.x, q.y, q.z, q.w);
    _euler = angles;
    _eulerOrder = order;
    _eulerDirty = false;
}

Vec3f
SFRotation::operator *(const Vec3f &v) const
{
    return _quaternion * v;
}
