/*
 * Scene.h
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#ifndef _SCENE_H
#define _SCENE_H

#ifndef _STDIO_H
#include <stdio.h>
#endif

#ifndef __gl_h
#include <GL/gl.h>
#endif

#ifndef _ARRAY_H
#include "Array.h"
#endif

#ifndef _MAP_H
#include "Map.h"
#endif

#ifndef _STACK_H
#include "Stack.h"
#endif

#ifndef _NODE_LIST_H
#include "NodeList.h"
#endif

#ifndef _DUNE_STRING_H
#include "String.h"
#endif

#ifndef _URL_H
#include "URL.h"
#endif

#define MAX_PATH_LEN 1024

class FieldValue;
class Node;
class Proto;
class Command;
class Path;
class FontInfo;
class NodeViewpoint;
class SFVec3f;
class SceneView;
class Hint;

typedef Stack<Command *>	CommandStack;

typedef Map<String, Node *>	NodeMap;
typedef Map<String, int>	StringMap;
typedef Map<String, Proto *>	ProtoMap;

typedef enum {
    TM_TRANSLATE,
    TM_ROTATE,
    TM_SCALE
} TransformMode;

class Interpolator;

class Scene {
public:
			Scene();
		       ~Scene();
    void		def(const char *nodeName, Node *value);
    Node	       *use(const char *nodeName);
    void		undef(const char *nodeName);

    int			addSymbol(String s);
    const String       &getSymbol(int id) const;

    void		setRoot(Node *root)	{ _root = root; }
    Node	       *getRoot() const			{ return _root; }
    void		setNodes(NodeList *nodes);

    const char         *parse(FILE *f);
    Node	       *createNode(const char *nodeType);
    void		write(FILE *f, const char *url, bool tempSave);

    void		addProto(String name, Proto *value);
    Proto	       *getProto(String name);

    static bool		validRoute(Node *src, int eventOut, Node *dst, int eventIn);
    void		addRoute(Node *src, int eventOut, Node *dest, int eventIn);
    void		deleteRoute(Node *src, int eventOut, Node *dest, int eventIn);

    int			getFieldType(String name);
    static const char  *getFieldTypeName(int type);

    void		execute(Command *cmd);

    void		add(Command *change);

    void		undo();
    void		redo();

    int			canUndo() { return !_undoStack.empty(); }
    int			canRedo() { return !_redoStack.empty(); }

    void		errorf(const char *fmt, ...);

    void		invalidNode(const char *nodeName);
    void		invalidField(const char *node, const char *field);

    void		backupField(Node *node, int field);
    void		setField(Node *node, int field, FieldValue *value);
    Interpolator       *findUpstreamInterpolator(Node *node, int field) const;

    void		drawScene(bool pick = false, int x = 0, int y = 0);
    Path	       *pick(int x, int y);
    void		drawHandles();
    void		drawHandlesRec(Node *node) const;

    void		setSelectedHandle(int handle)
			{ _selectedHandle = handle; }
    int			getSelectedHandle() const
			{ return _selectedHandle; }

    void		setTransformMode(TransformMode transformMode)
			{ _transformMode = transformMode; }
    TransformMode	getTransformMode()
			{ return _transformMode; }

    void		transform(const Path *path);
    void		projectPoint(float x, float y, float z,
				     float *wx, float *wy, float *wz);
    void		unProjectPoint(float wx, float wy, float wz,
				     float *x, float *y, float *z);
    int			allocateLight();
    int			freeLight();
    void		enableHeadlight();
    Path	       *processHits(unsigned hits, GLuint *pickBuffer);

    void		addViewpoint(Node *viewpoint);
    void		addFog(Node *fog);
    void		addBackground(Node *background);
    void		addTimeSensor(Node *timeSensor);

    void		setHeadlight(int headlight)
			{ _headlight = headlight; }

    void		moveCamera(float dx, float dy, float dz);
    void		turnCamera(float x, float y, float z, float ang);
    void		orbitCamera(float dtheta, float dphi);

    NodeViewpoint      *getCamera() const;

    void		applyCamera(void);

    void		addNode(Node *node);
    void		removeNode(Node *node);
    
    String		generateUniqueNodeName(Node *node);

    const NodeList     *getNodes() const { return &_nodes; }

    const Path	       *getSelection() const { return _selection; }
    void		setSelection(const Path *selection);
    void		setSelection(Node *node);

    void		setURL(const char *url) { _URL = url; }
    const char	       *getURL() const { return _URL; }
    const char	       *getNewURL() const { return _newURL; }
    bool		isTempSave() const { return _tempSave; }

    void		setPath(const char *path) { _path = path; }
    const char	       *getPath() const { return _path; }

    bool		isModified() const;
    bool		isRunning() const { return _running; }
    bool		isRecording() const { return _recording; }
    void		setRecording(bool recording) { _recording = recording; }

    void		start();
    void		stop();
    void		updateTime();

    void		OnFieldChange(Node *node, int field, int index = -1);
    void		OnAddNode(Node *node, Node *dest, int field);
    void		OnRemoveNode(Node *node, Node *src, int field);

    void		AddView(SceneView *view);
    void		RemoveView(SceneView *view);
    void		UpdateViews(SceneView *sender, int type,
				    Hint *hint = NULL);

    void		DeleteSelected();

    int			OnDragOver(Node *src, Node *srcParent, int srcField, 
				   Node *dest, int destField, int modifiers);
    int			OnDrop(Node *src, Node *srcParent, int srcField, 
			       Node *dest, int destField, int effect);

    bool		Download(const URL &url, String *path);
    FontInfo           *LoadGLFont(const char *fontName, const char *style);

private:
    NodeMap		_nodeMap;
    Node	       *_root;
    StringMap		_symbols;
    Array<String>	_symbolList;
    ProtoMap		_protos;
    StringMap		_fieldTypeMap;

    CommandStack	_undoStack;
    CommandStack	_redoStack;

    int			_numLights;
    NodeList		_viewpoints;
    NodeList		_fogs;
    NodeList		_backgrounds;
    NodeList		_timeSensors;
    Stack<Node *>	_fogStack;
    int			_headlight;

    NodeList		_nodes;

    const Path	       *_selection;
    int			_selectedHandle;
    TransformMode	_transformMode;

    String		_URL;
    String		_newURL;
    bool		_tempSave;

    int		        _numSymbols;

    bool		_running;
    bool		_recording;

    Command	       *_unmodified;

    NodeViewpoint      *_defaultViewpoint;
    NodeViewpoint      *_currentViewpoint;

    List<SceneView *>	_views;

    String		_compileErrors;
    Array<FontInfo *>	_fonts;

    String		_path;
};

enum {
    UPDATE_ALL = 0,
    UPDATE_SELECTION,
    UPDATE_FIELD,
    UPDATE_ADD_NODE,
    UPDATE_REMOVE_NODE,
    UPDATE_ADD_ROUTE,
    UPDATE_DELETE_ROUTE,
    UPDATE_MODE,
    UPDATE_TIME
};

class Hint {
};

class FieldUpdate : public Hint {
public:
                    FieldUpdate(Node *n, int f, int i = -1)
                    { node = n; field = f; index = i; }

    Node           *node;
    int             field;
    int             index;
};

class NodeUpdate : public Hint {
public:
                    NodeUpdate(Node *n, Node *p, int f)
                    { node = n; parent = p, field = f; }

    Node           *node;
    Node           *parent;
    int             field;
};

class RouteUpdate : public Hint {
public:
                    RouteUpdate(Node *s, int out, Node *d, int in)
                    { src = s; eventOut = out; dest = d; eventIn = in; }

    Node           *src;
    Node           *dest;
    int             eventOut;
    int             eventIn;
};

#endif // _SCENE_H
