/*
 * browser.c - win32 interface to web browser
 *
 * Copyright (C) 2000 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <swt.h>
#include <windows.h>

#define STARTUP_TIMEOUT	10000

typedef struct SBrowser {
    STABLE			prefs;
    char		       *command;
    char		       *application;
    int				useRemote;
    char		       *remoteCommand;
    char		       *topic;

    DWORD			DDEInst;
    HSZ				ourService;
    HSZ				theirService;
    HSZ				topicSZ;
    HSZ				exitSZ;
    HANDLE			hProcess;
} SBrowser;

static int	browserLaunch(SBROWSER browser, const char *path, SWND wnd);
static int	browserConnect(SBROWSER browser);
static int	browserGoTo(SBROWSER browser, const char *path);
static int	bringToFront(SBROWSER browser);
static char    *fmtMsg(LPCTSTR lpszFormat, ...);
static int	getShellOpenCommand(SBROWSER browser, const char *extension);
static void	saveBrowserPreferences(SBROWSER browser);

/*******************/
/* browser preview */
/*******************/

extern SBROWSER
swBrowserInit(STABLE prefs)
{
    SBrowser	*browser = malloc(sizeof(SBrowser));
    browser->prefs = prefs;
    browser->command = strdup(swGetPreference(prefs, "PreviewCommand", ""));
    browser->useRemote = swGetIntPreference(prefs, "PreviewUseRemote", TRUE);
    browser->remoteCommand = strdup(swGetPreference(prefs, "PreviewRemoteCommand",
				    ""));
    browser->application = strdup(swGetPreference(prefs, "PreviewApplication", ""));
    browser->topic = strdup(swGetPreference(prefs, "PreviewTopic", ""));

    if (!browser->command[0]) {
	getShellOpenCommand(browser, ".wrl");
    }

    browser->hProcess = NULL;
    browser->DDEInst = 0L;

    return browser;
}

extern void
swBrowserGetSettings(SBROWSER browser, const char **command, int *useRemote,
		     const char **remoteCommand, const char **application,
		     const char **topic)
{
    *command = browser->command;
    *useRemote = browser->useRemote;
    *remoteCommand = browser->remoteCommand;
    *application = browser->application;
    *topic = browser->topic;
}

extern void
swBrowserSetSettings(SBROWSER browser, const char *command, int useRemote,
		     const char *remoteCommand, const char *application,
		     const char *topic)
{
    free(browser->command);
    free(browser->remoteCommand);
    free(browser->application);
    free(browser->topic);

    browser->command = strdup(command);
    browser->useRemote = useRemote;
    browser->remoteCommand = strdup(remoteCommand);
    browser->application = strdup(application);
    browser->topic = strdup(topic);

    saveBrowserPreferences(browser);
}

static void
saveBrowserPreferences(SBROWSER browser)
{
    swSetPreference(browser->prefs, "PreviewCommand", browser->command);
    swSetIntPreference(browser->prefs, "PreviewUseRemote", browser->useRemote);
    swSetPreference(browser->prefs, "PreviewRemoteCommand", browser->remoteCommand);
    swSetPreference(browser->prefs, "PreviewApplication", browser->application);
    swSetPreference(browser->prefs, "PreviewTopic", browser->topic);
}

extern void
swBrowserPreview(SBROWSER browser, const char *path, SWND wnd)
{
    if (browser->useRemote) {
	if (!browserGoTo(browser, path)) {
	    if (browserLaunch(browser, NULL, wnd)) {
		browserGoTo(browser, path);
	    }
	}
    } else {
	browserLaunch(browser, path, wnd);
    }
}

extern void
swBrowserShutdown(SBROWSER browser)
{
    saveBrowserPreferences(browser);
    if (browser->hProcess) {
	TerminateProcess(browser->hProcess, 1);
	CloseHandle(browser->hProcess);
    }
    free(browser->command);
    free(browser->remoteCommand);
    free(browser->application);
    free(browser->topic);

    if(browser->DDEInst != 0) {
        DdeNameService( browser->DDEInst, NULL, NULL, DNS_UNREGISTER);
        DdeFreeStringHandle( browser->DDEInst, browser->theirService);
        DdeFreeStringHandle( browser->DDEInst, browser->ourService);
        DdeFreeStringHandle( browser->DDEInst, browser->topicSZ);
        DdeUninitialize( browser->DDEInst );
    }
    free(browser);
}

static char *
fmtMsg(LPCTSTR lpszFormat, ...)
{
    char    *buf;
    LPTSTR lpszTemp;

    // format message into temporary buffer lpszTemp
    va_list argList;
    va_start(argList, lpszFormat);

    FormatMessage(FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ALLOCATE_BUFFER,
		  lpszFormat, 0, 0, (LPTSTR)&lpszTemp, 0, &argList);

    // assign lpszTemp into the resulting string and free the temporary
    buf = strdup(lpszTemp);
    LocalFree(lpszTemp);
    va_end(argList);
    return buf;
}

static int browserLaunch(SBROWSER browser, const char *path, SWND wnd)
{
    STARTUPINFO	 startInfo;
    PROCESS_INFORMATION processInfo;
    int	rc;
    char       *s, *cmdline;

    if (path && path[0]) {
	cmdline = fmtMsg(browser->command, path);
    } else {
	cmdline = strdup(browser->command);
	if (cmdline[0] == '"') {
	    s = strchr(cmdline+1, '"');
	    if (s) (*(s+1)) = 0;
	} else {
	    s = strchr(cmdline, ' ');
	    if (s) *s = 0;
	}
    }

    memset(&startInfo, 0, sizeof(startInfo));

    startInfo.cb = sizeof(STARTUPINFO);
    startInfo.dwFlags = STARTF_USESHOWWINDOW;
    startInfo.wShowWindow = SW_SHOWNA;

    if (CreateProcess(NULL, cmdline, NULL, NULL, FALSE, 0, NULL, NULL,
		      &startInfo, &processInfo) != 0) {
	WaitForInputIdle( processInfo.hProcess, STARTUP_TIMEOUT );
	CloseHandle(processInfo.hThread);
	browser->hProcess = processInfo.hProcess;
	rc = TRUE;
    } else {
	rc = FALSE;
    }
    free(cmdline);
    return rc;
}

static HDDEDATA DDECallback(
    UINT type,		/* transaction type */
    UINT fmt,		/* clipboard data format    */
    HCONV hconv,	/* handle of conversation   */
    HSZ hsz1,		/* handle of string */
    HSZ hsz2,		/* handle of string */
    HDDEDATA hData,	/* handle of global memory object   */
    DWORD dwData1,	/* transaction-specific data    */
    DWORD dwData2	/* transaction-specific data    */ )
{
    return (HDDEDATA) 1;
}

static int browserGoTo(SBROWSER browser, const char *path)
{
    char       *buf;

    HSZ		data;
    HCONV	conv;
    HDDEDATA	res;
    int		timeout = 120;
    int		rc = FALSE;

    if (browser->DDEInst == 0) {
	if (!browserConnect(browser)) return FALSE;
    }
    conv = DdeConnect( browser->DDEInst, browser->theirService,
		       browser->topicSZ, NULL );
    if( conv != NULL) {
	buf = fmtMsg( browser->remoteCommand, path );
	data = DdeCreateStringHandle( browser->DDEInst, buf, CP_WINANSI );

	res = DdeClientTransaction( NULL, 0, conv, data, CF_TEXT, XTYP_REQUEST,
				    TIMEOUT_ASYNC, NULL);

	DdeDisconnect( conv);
	DdeFreeStringHandle( browser->DDEInst, data);
	free(buf);
	bringToFront(browser);
	rc = TRUE;
    }
    return rc;
}

LONG queryValue(HKEY key, const char *subKey, char **result)
{
    LONG            size, rc;

    rc = RegQueryValue(key, subKey, NULL, &size);
    if (rc == ERROR_SUCCESS) {
        *result = (char *) malloc(size + 1);
        RegQueryValue(key, subKey, *result, &size);
    } else {
	*result = strdup("");
    }
    return rc;
}

static int
getShellOpenCommand(SBROWSER browser, const char *extension)
{
    HKEY		key;
    char	       *info;

    if (queryValue(HKEY_CLASSES_ROOT, extension, &info) != ERROR_SUCCESS) {
        return FALSE;
    }

    if (RegOpenKey(HKEY_CLASSES_ROOT, info, &key) != ERROR_SUCCESS) {
	free(info);
        return FALSE;
    }

    free(browser->command);
    free(browser->remoteCommand);
    free(browser->application);
    free(browser->topic);
    queryValue(key, "Shell\\Open\\command", &browser->command);
    queryValue(key, "Shell\\Open\\ddeexec", &browser->remoteCommand);
    browser->useRemote = browser->remoteCommand[0];
    queryValue(key, "Shell\\Open\\ddeexec\\Application", &browser->application);
    queryValue(key, "Shell\\Open\\ddeexec\\Topic", &browser->topic);

    RegCloseKey(key);
    free(info);
    return TRUE;
}

static int
browserConnect(SBROWSER browser)
{
    int		rc = FALSE;

    UINT initRes = DdeInitialize( &browser->DDEInst,
				  (PFNCALLBACK) DDECallback,
	 APPCLASS_STANDARD | CBF_FAIL_CONNECTIONS | CBF_SKIP_DISCONNECTS,
				  0 );

    if (initRes == DMLERR_NO_ERROR) {
        browser->theirService = DdeCreateStringHandle( browser->DDEInst, browser->application, CP_WINANSI );
        browser->ourService = DdeCreateStringHandle( browser->DDEInst, "Dune", CP_WINANSI );
        browser->topicSZ = DdeCreateStringHandle( browser->DDEInst, browser->topic, CP_WINANSI );

        DdeNameService( browser->DDEInst, browser->ourService, NULL, DNS_REGISTER);
	rc = TRUE;
    }
    return rc;
}

static int bringToFront(SBROWSER browser)
{
    HSZ         data, activate;
    HCONV       conv;
    int         timeout = 120;
    int		rc = FALSE;

    activate = DdeCreateStringHandle( browser->DDEInst, "WWW_Activate", CP_WINANSI );
    conv = DdeConnect( browser->DDEInst, browser->theirService, activate, NULL );
    if( conv != NULL) {
        data = DdeCreateStringHandle( browser->DDEInst, "0xFFFFFFFF", CP_WINANSI );

        DdeClientTransaction( NULL, 0, conv, data, CF_TEXT, XTYP_REQUEST,
                        TIMEOUT_ASYNC, NULL);

        DdeFreeStringHandle( browser->DDEInst, data);
        DdeDisconnect( conv);
        rc = TRUE;
    }
    DdeFreeStringHandle( browser->DDEInst, activate);
    return rc;
}
